__name__ = "deep_merge"
__version__ = "0.0.4"
__author__ = "Aaron Halfaker"
__author_email__ = "aaron.halfaker@gmail.com"
__description__ = "A simple utility for merging python dictionaries."
__url__ = "https://github.com/halfak/deep_merge"
__license__ = "MIT"



import copy


def append_lists(l1, l2, **kwargs):
    """
    Appends the values in one list on the end of another list
    """
    l1.extend(l2)
    return l1


def overwrite(v1, v2, **kwargs):
    """
    Completely overwrites one value with another.
    """
    return copy.deepcopy(v2)


def merge_dicts(d1, d2, merge_lists=overwrite,
                merge_ints=overwrite, merge_floats=overwrite,
                merge_strings=overwrite, merge_other=overwrite):
    """
    Recursively merges values from d2 into d1.
    """
    kwargs = {'merge_lists': merge_lists,
              'merge_ints': merge_ints,
              'merge_floats': merge_floats,
              'merge_strings': merge_strings,
              'merge_other': merge_other}
    for key in d2:
        if key in d1:
            if isinstance(d1[key], dict) and isinstance(d2[key], dict):
                d1[key] = merge_dicts(d1[key], d2[key], **kwargs)
            elif isinstance(d1[key], list) and isinstance(d2[key], list):
                d1[key] = merge_lists(d1[key], d2[key], **kwargs)
            elif isinstance(d1[key], int) and isinstance(d2[key], int):
                d1[key] = merge_ints(d1[key], d2[key], **kwargs)
            elif isinstance(d1[key], float) and isinstance(d2[key], float):
                d1[key] = merge_ints(d1[key], d2[key], **kwargs)
            elif isinstance(d1[key], str) and isinstance(d2[key], str):
                d1[key] = merge_strings(d1[key], d2[key], **kwargs)
            else:
                d1[key] = merge_other(d1[key], d2[key], **kwargs)
        else:
            d1[key] = overwrite(None, d2[key])

    return d1


def merge(*dicts, merge_dicts=merge_dicts, merge_lists=overwrite,
          merge_ints=overwrite, merge_floats=overwrite,
          merge_strings=overwrite, merge_other=overwrite):
    """
    Recursively merges dictionaries and the datastructures they contain.

    :Parameters:
        *dicts : `dict`
            Dictionaries to be merged.  Items that appear last will take higher
            precedence when merging.
        merge_dicts : `func`
            The function to apply when merging dictionaries.
        merge_lists : `func`
            The function to apply when merging lists.
        merge_ints : `func`
            The function to apply when merging integers.
        merge_floats : `func`
            The function to apply when merging floats.
        merge_strings : `func`
            The function to apply when merging strings.
        merge_other : `func`
            The function to apply when merging other types or types that do not
            match.
    """
    for param in dicts:
        if not isinstance(param, dict):
            raise TypeError("{0} is not a dict".format(param))

    d = dicts[0]
    for d_update in dicts[1:]:
        d = merge_dicts(d, d_update,
                        merge_lists=merge_lists,
                        merge_ints=merge_ints, merge_floats=merge_floats,
                        merge_strings=merge_strings, merge_other=merge_other)

    return d



import os
from distutils.core import setup

from setuptools import find_packages

about_path = os.path.join(os.path.dirname(__file__), "deep_merge/about.py")
exec(compile(open(about_path).read(), about_path, "exec"))


setup(
    name=__name__,  # noqa
    version=__version__,  # noqa
    author=__author__,  # noqa
    author_email=__author_email__,  # noqa
    description=__description__,  # noqa
    url=__url__,  # noqa
    license=__license__,  # noqa
    packages=find_packages(),
    long_description=open('README.md').read(),
    test_suite='nose.collector',
    classifiers=[
        "Programming Language :: Python",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)



from ..merge import merge


def test_default_behaviors():
    a = {'foo': 5}
    b = {'foo': False}
    assert merge(a, b) == {'foo': False}

    a = {'foo': 5}
    b = {'foo': 10}
    assert merge(a, b, merge_ints=lambda a, b, **kwargs: a+b) == {'foo': 15}

    a = {'foo': 5}
    b = {'bar': 10}
    assert merge(a, b) == {'foo': 5, 'bar': 10}


def test_big_merge():
    a = {
        'foo': {
            'bar': {
                'foo': 5,
                'bar': 7
            }
        },
        'l': [1, 5, 'foo']
    }
    b = {
        'foo': {
            'bar': {
                'foo': 6
            }
        },
        'bar': 5,
        's': "I'm a string",
        'l': [1, 2, 3],
        'herpa_derpa': 5
    }
    c = {
        'foo': {
            'bar': {
                'foobar': 10
            }
        },
        'herpa_derpa': None
    }
    expected = {
        'foo': {
            'bar': {
                'foo': 6,
                'bar': 7,
                'foobar': 10
            }
        },
        'bar': 5,
        's': "I'm a string",
        'l': [1, 5, 'foo', 1, 2, 3],
        'herpa_derpa': None
    }
    assert merge(a, b, c) == expected


def test_no_references():
    """
    Ensure that read-only parameters aren't being modified.
    """
    d1 = {}
    d2 = {"key1": {"key1a": "1"}, "key2": [3, 4]}
    d3 = {"key1": {"key1a": "2"}, "key2": [4, 5]}

    output = merge(d1, d2, d3)
    # Is correct.
    assert output == {"key1": {"key1a": "2"}, "key2": [3, 4, 4, 5]}
    # Destructive to parameter 1.
    assert d1 == {"key1": {"key1a": "2"}, "key2": [3, 4, 4, 5]}
    # Not destructive to parameter 2.
    assert d2 == {"key1": {"key1a": "1"}, "key2": [3, 4]}
